@@ -38,8 +38,9 @@ import txaio
 
 from autobahn.util import ObservableMixin
 from autobahn.websocket.util import parse_url
-from autobahn.wamp.types import ComponentConfig
+from autobahn.wamp.types import ComponentConfig, SubscribeOptions, RegisterOptions
 from autobahn.wamp.exception import SessionNotReady
+from autobahn.wamp.auth import create_authenticator
 
 
 __all__ = ('Connection')
@@ -116,7 +117,11 @@ def _create_transport(index, transport, check_native_endpoint=None):
     if type(transport) != dict:
         raise ValueError('invalid type {} for transport configuration - must be a dict'.format(type(transport)))
 
-    valid_transport_keys = ['type', 'url', 'endpoint', 'serializer', 'serializers']
+    valid_transport_keys = [
+        'type', 'url', 'endpoint', 'serializer', 'serializers', 'options',
+        'max_retries', 'max_retry_delay', 'initial_retry_delay',
+        'retry_delay_growth', 'retry_delay_jitter',
+    ]
     for k in transport.keys():
         if k not in valid_transport_keys:
             raise ValueError(
@@ -131,6 +136,14 @@ def _create_transport(index, transport, check_native_endpoint=None):
     else:
         transport['type'] = 'websocket'
 
+    options = dict()
+    if 'options' in transport:
+        options = transport['options']
+        if not isinstance(options, dict):
+            raise ValueError(
+                'options must be a dict, not {}'.format(type(options))
+            )
+
     if kind == 'websocket':
         for key in ['url']:
             if key not in transport:
@@ -143,7 +156,7 @@ def _create_transport(index, transport, check_native_endpoint=None):
                 'type': 'tcp',
                 'host': host,
                 'port': port,
-                'tls': False if not is_secure else dict(hostname=host),
+                'tls': is_secure,
             }
         else:
             # note: we're avoiding mutating the incoming "configuration"
@@ -200,6 +213,7 @@ def _create_transport(index, transport, check_native_endpoint=None):
         url=transport['url'],
         endpoint=endpoint_config,
         serializers=serializer_config,
+        options=options,
         **kw
     )
 
@@ -214,7 +228,8 @@ class _Transport(object):
                  max_retry_delay=300,
                  initial_retry_delay=1.5,
                  retry_delay_growth=1.5,
-                 retry_delay_jitter=0.1):
+                 retry_delay_jitter=0.1,
+                 options=dict()):
         """
         """
         self.idx = idx
@@ -222,6 +237,7 @@ class _Transport(object):
         self.type = kind
         self.url = url
         self.endpoint = endpoint
+        self.options = options
 
         self.serializers = serializers
         if self.type == 'rawsocket' and len(serializers) != 1:
@@ -300,7 +316,51 @@ class Component(ObservableMixin):
     The factory of the session we will instantiate.
     """
 
-    def __init__(self, main=None, transports=None, config=None, realm=u'default', extra=None):
+    def subscribe(self, topic, options=None):
+        """
+        A decorator as a shortcut for subscribing during on-join
+
+        For example::
+
+            @component.subscribe(
+                u"some.topic",
+                options=SubscribeOptions(match=u'prefix'),
+            )
+            def topic(*args, **kw):
+                print("some.topic({}, {}): event received".format(args, kw))
+        """
+        assert options is None or isinstance(options, SubscribeOptions)
+
+        def decorator(fn):
+
+            def do_subscription(session, details):
+                return session.subscribe(fn, topic=topic, options=options)
+            self.on('join', do_subscription)
+        return decorator
+
+    def register(self, uri, options=None):
+        """
+        A decorator as a shortcut for registering during on-join
+
+        For example::
+
+            @component.register(
+                u"com.example.add",
+                options=RegisterOptions(invoke='round_robin'),
+            )
+            def add(*args, **kw):
+                print("add({}, {}): event received".format(args, kw))
+        """
+        assert options is None or isinstance(options, RegisterOptions)
+
+        def decorator(fn):
+
+            def do_registration(session, details):
+                return session.register(fn, procedure=uri, options=options)
+            self.on('join', do_registration)
+        return decorator
+
+    def __init__(self, main=None, transports=None, config=None, realm=u'default', extra=None, authentication=None):
         """
         :param main: After a transport has been connected and a session
             has been established and joined to a realm, this (async)
@@ -336,6 +396,9 @@ class Component(ObservableMixin):
 
         :param realm: the realm to join
         :type realm: unicode
+
+        :param authentication: configuration of authenticators
+        :type authentication: dict mapping auth_type to dict
         """
         self.set_valid_events(
             [
@@ -382,6 +445,9 @@ class Component(ObservableMixin):
                 _create_transport(idx, transport, self._check_native_endpoint)
             )
 
+        # XXX should have some checkconfig support
+        self._authentication = authentication or {}
+
         self._realm = realm
         self._extra = extra
 
@@ -414,6 +480,10 @@ class Component(ObservableMixin):
             cfg = ComponentConfig(self._realm, self._extra)
             try:
                 session = self.session_factory(cfg)
+                for auth_name, auth_config in self._authentication.items():
+                    authenticator = create_authenticator(auth_name, **auth_config)
+                    session.add_authenticator(authenticator)
+
             except Exception as e:
                 # couldn't instantiate session calls, which is fatal.
                 # let the reconnection logic deal with that
@@ -436,7 +506,7 @@ class Component(ObservableMixin):
                         "session leaving '{details.reason}'",
                         details=details,
                     )
-                    if self._entry:
+                    if self._entry and not txaio.is_called(done):
                         txaio.resolve(done, None)
                 session.on('leave', on_leave)
 
@@ -444,6 +514,7 @@ class Component(ObservableMixin):
                 # it completely (i.e. until its Deferred fires) and
                 # then disconnect this session
                 def on_join(session, details):
+                    transport.connect_sucesses += 1
                     self.log.debug("session on_join: {details}", details=details)
                     d = txaio.as_future(self._entry, reactor, session)
 
@@ -462,6 +533,7 @@ class Component(ObservableMixin):
                     def main_error(err):
                         self.log.debug("main_error: {err}", err=err)
                         txaio.reject(done, err)
+                        session.disconnect()
                     txaio.add_callbacks(d, main_success, main_error)
                 if self._entry is not None:
                     session.on('join', on_join)
@@ -509,7 +581,7 @@ class Component(ObservableMixin):
         def on_connect_failure(err):
             transport.connect_failures += 1
             # failed to establish a connection in the first place
-            done.errback(err)
+            txaio.reject(done, err)
 
         txaio.add_callbacks(d, on_connect_sucess, None)
         txaio.add_callbacks(d, None, on_connect_failure)
@@ -601,7 +673,7 @@ def _run(reactor, components):
     def component_start(comp):
         # the future from start() errbacks if we fail, or callbacks
         # when the component is considered "done" (so maybe never)
-        d = comp.start(reactor)
+        d = txaio.as_future(comp.start, reactor)
         txaio.add_callbacks(
             d,
             partial(component_success, comp),
