@@ -29,6 +29,7 @@ from __future__ import absolute_import
 import six
 import txaio
 import inspect
+from functools import reduce
 
 from autobahn import wamp
 from autobahn.util import public, IdGenerator, ObservableMixin
@@ -38,9 +39,8 @@ from autobahn.wamp import types
 from autobahn.wamp import role
 from autobahn.wamp import exception
 from autobahn.wamp.exception import ApplicationError, ProtocolError, SessionNotReady, SerializationError
-from autobahn.wamp.interfaces import ISession  # noqa
-from autobahn.wamp.types import SessionDetails, CloseDetails
-from autobahn.wamp.cryptobox import EncryptedPayload
+from autobahn.wamp.interfaces import ISession, IPayloadCodec, IAuthenticator  # noqa
+from autobahn.wamp.types import SessionDetails, CloseDetails, EncodedPayload
 from autobahn.wamp.request import \
     Publication, \
     Subscription, \
@@ -67,6 +67,8 @@ class BaseSession(ObservableMixin):
     This class implements :class:`autobahn.wamp.interfaces.ISession`.
     """
 
+    log = txaio.make_logger()
+
     def __init__(self):
         """
 
@@ -100,8 +102,8 @@ class BaseSession(ObservableMixin):
         self._authmethod = None
         self._authprovider = None
 
-        # end-to-end encryption keyring
-        self._keyring = None
+        # payload transparency codec
+        self._payload_codec = None
 
         # generator for WAMP request IDs
         self._request_id_gen = IdGenerator()
@@ -125,15 +127,16 @@ class BaseSession(ObservableMixin):
 
         :param request_type: The request type this WAMP error message is for.
         :type request_type: int
+
         :param request: The request ID this WAMP error message is for.
         :type request: int
+
         :param exc: The exception.
         :type exc: Instance of :class:`Exception` or subclass thereof.
+
         :param tb: Optional traceback. If present, it'll be included with the WAMP error message.
         :type tb: list or None
         """
-        assert(enc_algo is None or enc_algo == message.PAYLOAD_ENC_CRYPTO_BOX)
-
         args = None
         if hasattr(exc, 'args'):
             args = list(exc.args)  # make sure tuples are made into lists
@@ -156,18 +159,18 @@ class BaseSession(ObservableMixin):
             else:
                 error = u"wamp.error.runtime_error"
 
-        encrypted_payload = None
-        if self._keyring:
-            encrypted_payload = self._keyring.encrypt(False, error, args, kwargs)
+        encoded_payload = None
+        if self._payload_codec:
+            encoded_payload = self._payload_codec.encode(False, error, args, kwargs)
 
-        if encrypted_payload:
+        if encoded_payload:
             msg = message.Error(request_type,
                                 request,
                                 error,
-                                payload=encrypted_payload.payload,
-                                enc_algo=encrypted_payload.algo,
-                                enc_key=encrypted_payload.pkey,
-                                enc_serializer=encrypted_payload.serializer)
+                                payload=encoded_payload.payload,
+                                enc_algo=encoded_payload.enc_algo,
+                                enc_key=encoded_payload.enc_key,
+                                enc_serializer=encoded_payload.enc_serializer)
         else:
             msg = message.Error(request_type,
                                 request,
@@ -192,16 +195,16 @@ class BaseSession(ObservableMixin):
         exc = None
         enc_err = None
 
-        if msg.enc_algo == message.PAYLOAD_ENC_CRYPTO_BOX:
+        if msg.enc_algo:
 
-            if not self._keyring:
-                log_msg = u"received encrypted payload, but no keyring active"
+            if not self._payload_codec:
+                log_msg = u"received encoded payload, but no payload codec active"
                 self.log.warn(log_msg)
-                enc_err = ApplicationError(ApplicationError.ENC_NO_KEYRING_ACTIVE, log_msg, enc_algo=msg.enc_algo)
+                enc_err = ApplicationError(ApplicationError.ENC_NO_PAYLOAD_CODEC, log_msg, enc_algo=msg.enc_algo)
             else:
                 try:
-                    encrypted_payload = EncryptedPayload(msg.enc_algo, msg.enc_key, msg.enc_serializer, msg.payload)
-                    decrypted_error, msg.args, msg.kwargs = self._keyring.decrypt(True, msg.error, encrypted_payload)
+                    encoded_payload = EncodedPayload(msg.payload, msg.enc_algo, msg.enc_serializer, msg.enc_key)
+                    decrypted_error, msg.args, msg.kwargs = self._payload_codec.decode(True, msg.error, encoded_payload)
                 except Exception as e:
                     self.log.warn("failed to decrypt application payload 1: {err}", err=e)
                     enc_err = ApplicationError(
@@ -307,11 +310,19 @@ class ApplicationSession(BaseSession):
         self._invocations = {}
 
     @public
-    def set_keyring(self, keyring):
+    def set_payload_codec(self, payload_codec):
         """
-        Implements :func:`autobahn.wamp.interfaces.ISession.set_keyring`
+        Implements :func:`autobahn.wamp.interfaces.ISession.set_payload_codec`
         """
-        self._keyring = keyring
+        assert(payload_codec is None or isinstance(payload_codec, IPayloadCodec))
+        self._payload_codec = payload_codec
+
+    @public
+    def get_payload_codec(self):
+        """
+        Implements :func:`autobahn.wamp.interfaces.ISession.get_payload_codec`
+        """
+        return self._payload_codec
 
     @public
     def onOpen(self, transport):
@@ -408,10 +419,8 @@ class ApplicationSession(BaseSession):
         """
         Implements :func:`autobahn.wamp.interfaces.ISession.onUserError`
         """
-        if isinstance(fail.value, exception.ApplicationError):
-            # silence on errors raised explicitly from the app
-            # previous code: self.log.error(fail.value.error_message())
-            pass
+        if False and isinstance(fail.value, exception.ApplicationError):
+            self.log.error(fail.value.error_message())
         else:
             self.log.error(
                 u'{msg}: {traceback}',
@@ -597,19 +606,22 @@ class ApplicationSession(BaseSession):
                         handler = subscription.handler
                         topic = msg.topic or subscription.topic
 
-                        if msg.enc_algo == message.PAYLOAD_ENC_CRYPTO_BOX:
+                        if msg.enc_algo:
                             # FIXME: behavior in error cases (no keyring, decrypt issues, URI mismatch, ..)
-                            if not self._keyring:
-                                self.log.warn("received encrypted payload, but no keyring active - ignoring encrypted payload!")
+                            if not self._payload_codec:
+                                self.log.warn("received encoded payload with enc_algo={enc_algo}, but no payload codec active - ignoring encoded payload!", enc_algo=msg.enc_algo)
+                                return
                             else:
                                 try:
-                                    encrypted_payload = EncryptedPayload(msg.enc_algo, msg.enc_key, msg.enc_serializer, msg.payload)
-                                    decrypted_topic, msg.args, msg.kwargs = self._keyring.decrypt(False, topic, encrypted_payload)
+                                    encoded_payload = EncodedPayload(msg.payload, msg.enc_algo, msg.enc_serializer, msg.enc_key)
+                                    decoded_topic, msg.args, msg.kwargs = self._payload_codec.decode(False, topic, encoded_payload)
                                 except Exception as e:
-                                    self.log.warn("failed to decrypt application payload: {error}", error=e)
+                                    self.log.warn("failed to decode application payload encoded with enc_algo={enc_algo}: {error}", error=e, enc_algo=msg.enc_algo)
+                                    return
                                 else:
-                                    if topic != decrypted_topic:
-                                        self.log.warn("envelope topic URI does not match encrypted one")
+                                    if topic != decoded_topic:
+                                        self.log.warn("envelope topic URI does not match encoded one")
+                                        return
 
                         invoke_args = (handler.obj,) if handler.obj else tuple()
                         if msg.args:
@@ -617,7 +629,7 @@ class ApplicationSession(BaseSession):
                         invoke_kwargs = msg.kwargs if msg.kwargs else dict()
 
                         if handler.details_arg:
-                            invoke_kwargs[handler.details_arg] = types.EventDetails(publication=msg.publication, publisher=msg.publisher, publisher_authid=msg.publisher_authid, publisher_authrole=msg.publisher_authrole, topic=topic, retained=msg.retained, enc_algo=msg.enc_algo)
+                            invoke_kwargs[handler.details_arg] = types.EventDetails(subscription, msg.publication, publisher=msg.publisher, publisher_authid=msg.publisher_authid, publisher_authrole=msg.publisher_authrole, topic=topic, retained=msg.retained, enc_algo=msg.enc_algo)
 
                         # FIXME: https://github.com/crossbario/autobahn-python/issues/764
                         def _success(_):
@@ -703,16 +715,16 @@ class ApplicationSession(BaseSession):
                     proc = call_request.procedure
                     enc_err = None
 
-                    if msg.enc_algo == message.PAYLOAD_ENC_CRYPTO_BOX:
+                    if msg.enc_algo:
 
-                        if not self._keyring:
-                            log_msg = u"received encrypted payload, but no keyring active"
+                        if not self._payload_codec:
+                            log_msg = u"received encoded payload, but no payload codec active"
                             self.log.warn(log_msg)
-                            enc_err = ApplicationError(ApplicationError.ENC_NO_KEYRING_ACTIVE, log_msg)
+                            enc_err = ApplicationError(ApplicationError.ENC_NO_PAYLOAD_CODEC, log_msg)
                         else:
                             try:
-                                encrypted_payload = EncryptedPayload(msg.enc_algo, msg.enc_key, msg.enc_serializer, msg.payload)
-                                decrypted_proc, msg.args, msg.kwargs = self._keyring.decrypt(True, proc, encrypted_payload)
+                                encoded_payload = EncodedPayload(msg.payload, msg.enc_algo, msg.enc_serializer, msg.enc_key)
+                                decrypted_proc, msg.args, msg.kwargs = self._payload_codec.decode(True, proc, encoded_payload)
                             except Exception as e:
                                 self.log.warn(
                                     "failed to decrypt application payload 1: {err}",
@@ -801,15 +813,15 @@ class ApplicationSession(BaseSession):
                         proc = msg.procedure or registration.procedure
                         enc_err = None
 
-                        if msg.enc_algo == message.PAYLOAD_ENC_CRYPTO_BOX:
-                            if not self._keyring:
+                        if msg.enc_algo:
+                            if not self._payload_codec:
                                 log_msg = u"received encrypted INVOCATION payload, but no keyring active"
                                 self.log.warn(log_msg)
-                                enc_err = ApplicationError(ApplicationError.ENC_NO_KEYRING_ACTIVE, log_msg)
+                                enc_err = ApplicationError(ApplicationError.ENC_NO_PAYLOAD_CODEC, log_msg)
                             else:
                                 try:
-                                    encrypted_payload = EncryptedPayload(msg.enc_algo, msg.enc_key, msg.enc_serializer, msg.payload)
-                                    decrypted_proc, msg.args, msg.kwargs = self._keyring.decrypt(False, proc, encrypted_payload)
+                                    encoded_payload = EncodedPayload(msg.payload, msg.enc_algo, msg.enc_serializer, msg.enc_key)
+                                    decrypted_proc, msg.args, msg.kwargs = self._payload_codec.decode(False, proc, encoded_payload)
                                 except Exception as e:
                                     self.log.warn(
                                         "failed to decrypt INVOCATION payload: {err}",
@@ -855,19 +867,19 @@ class ApplicationSession(BaseSession):
                                 if msg.receive_progress:
 
                                     def progress(*args, **kwargs):
-                                        encrypted_payload = None
-                                        if msg.enc_algo == message.PAYLOAD_ENC_CRYPTO_BOX:
-                                            if not self._keyring:
+                                        encoded_payload = None
+                                        if msg.enc_algo:
+                                            if not self._payload_codec:
                                                 raise Exception(u"trying to send encrypted payload, but no keyring active")
-                                            encrypted_payload = self._keyring.encrypt(False, proc, args, kwargs)
+                                            encoded_payload = self._payload_codec.encode(False, proc, args, kwargs)
 
-                                        if encrypted_payload:
+                                        if encoded_payload:
                                             progress_msg = message.Yield(msg.request,
-                                                                         payload=encrypted_payload.payload,
+                                                                         payload=encoded_payload.payload,
                                                                          progress=True,
-                                                                         enc_algo=encrypted_payload.algo,
-                                                                         enc_key=encrypted_payload.pkey,
-                                                                         enc_serializer=encrypted_payload.serializer)
+                                                                         enc_algo=encoded_payload.enc_algo,
+                                                                         enc_key=encoded_payload.enc_key,
+                                                                         enc_serializer=encoded_payload.enc_serializer)
                                         else:
                                             progress_msg = message.Yield(msg.request,
                                                                          args=args,
@@ -878,36 +890,36 @@ class ApplicationSession(BaseSession):
                                 else:
                                     progress = None
 
-                                invoke_kwargs[endpoint.details_arg] = types.CallDetails(progress, caller=msg.caller, caller_authid=msg.caller_authid, caller_authrole=msg.caller_authrole, procedure=proc, enc_algo=msg.enc_algo)
+                                invoke_kwargs[endpoint.details_arg] = types.CallDetails(registration, progress=progress, caller=msg.caller, caller_authid=msg.caller_authid, caller_authrole=msg.caller_authrole, procedure=proc, enc_algo=msg.enc_algo)
 
                             on_reply = txaio.as_future(endpoint.fn, *invoke_args, **invoke_kwargs)
 
                             def success(res):
                                 del self._invocations[msg.request]
 
-                                encrypted_payload = None
-                                if msg.enc_algo == message.PAYLOAD_ENC_CRYPTO_BOX:
-                                    if not self._keyring:
+                                encoded_payload = None
+                                if msg.enc_algo:
+                                    if not self._payload_codec:
                                         log_msg = u"trying to send encrypted payload, but no keyring active"
                                         self.log.warn(log_msg)
                                     else:
                                         try:
                                             if isinstance(res, types.CallResult):
-                                                encrypted_payload = self._keyring.encrypt(False, proc, res.results, res.kwresults)
+                                                encoded_payload = self._payload_codec.encode(False, proc, res.results, res.kwresults)
                                             else:
-                                                encrypted_payload = self._keyring.encrypt(False, proc, [res])
+                                                encoded_payload = self._payload_codec.encode(False, proc, [res])
                                         except Exception as e:
                                             self.log.warn(
                                                 "failed to encrypt application payload: {err}",
                                                 err=e,
                                             )
 
-                                if encrypted_payload:
+                                if encoded_payload:
                                     reply = message.Yield(msg.request,
-                                                          payload=encrypted_payload.payload,
-                                                          enc_algo=encrypted_payload.algo,
-                                                          enc_key=encrypted_payload.pkey,
-                                                          enc_serializer=encrypted_payload.serializer)
+                                                          payload=encoded_payload.payload,
+                                                          enc_algo=encoded_payload.enc_algo,
+                                                          enc_key=encoded_payload.enc_key,
+                                                          enc_serializer=encoded_payload.enc_serializer)
                                 else:
                                     if isinstance(res, types.CallResult):
                                         reply = message.Yield(msg.request,
@@ -1001,7 +1013,23 @@ class ApplicationSession(BaseSession):
 
             elif isinstance(msg, message.Unregistered):
 
-                if msg.request in self._unregister_reqs:
+                if msg.request == 0:
+                    # this is a forced un-register either from a call
+                    # to the wamp.* meta-api or the force_reregister
+                    # option
+                    try:
+                        reg = self._registrations[msg.registration]
+                    except KeyError:
+                        raise ProtocolError(
+                            "UNREGISTERED received for non-existant registration"
+                            " ID {0}".format(msg.registration)
+                        )
+                    self.log.info(
+                        u"Router unregistered procedure '{proc}' with ID {id}",
+                        proc=reg.procedure,
+                        id=msg.registration,
+                    )
+                elif msg.request in self._unregister_reqs:
 
                     # get and pop outstanding subscribe request
                     request = self._unregister_reqs.pop(msg.request)
@@ -1108,20 +1136,38 @@ class ApplicationSession(BaseSession):
         """
         Errback any still outstanding requests with exc.
         """
-        for requests in [self._publish_reqs,
-                         self._subscribe_reqs,
-                         self._unsubscribe_reqs,
-                         self._call_reqs,
-                         self._register_reqs,
-                         self._unregister_reqs]:
-            for request in requests.values():
-                self.log.info('cleaning up outstanding {request_type} request {request_id}, firing errback on user handler {request_on_reply}',
-                              request_on_reply=request.on_reply,
-                              request_id=request.request_id,
-                              request_type=request.__class__.__name__)
-                txaio.reject(request.on_reply, exc)
+        d = txaio.create_future_success(None)
+        all_requests = [
+            self._publish_reqs,
+            self._subscribe_reqs,
+            self._unsubscribe_reqs,
+            self._call_reqs,
+            self._register_reqs,
+            self._unregister_reqs
+        ]
+        outstanding = []
+        for requests in all_requests:
+            outstanding.extend(requests.values())
             requests.clear()
 
+        if outstanding:
+            self.log.info(
+                'Cancelling {count} outstanding requests',
+                count=len(outstanding),
+            )
+        for request in outstanding:
+            self.log.debug(
+                'cleaning up outstanding {request_type} request {request_id}, '
+                'firing errback on user handler {request_on_reply}',
+                request_on_reply=request.on_reply,
+                request_id=request.request_id,
+                request_type=request.__class__.__name__,
+            )
+            txaio.reject(request.on_reply, exc)
+            # wait for any async-ness in the error handlers for on_reply
+            txaio.add_callbacks(d, lambda _: request.on_reply, lambda _: request.on_reply)
+        return d
+
     @public
     def onLeave(self, details):
         """
@@ -1132,10 +1178,13 @@ class ApplicationSession(BaseSession):
 
         # fire ApplicationError on any currently outstanding requests
         exc = ApplicationError(details.reason, details.message)
-        self._errback_outstanding_requests(exc)
+        d = self._errback_outstanding_requests(exc)
 
-        if self._transport:
-            self.disconnect()
+        def disconnect(_):
+            if self._transport:
+                self.disconnect()
+        txaio.add_callbacks(d, disconnect, disconnect)
+        return d
 
     @public
     def leave(self, reason=None, message=None):
@@ -1184,26 +1233,26 @@ class ApplicationSession(BaseSession):
 
         request_id = self._request_id_gen.next()
 
-        encrypted_payload = None
-        if self._keyring:
-            encrypted_payload = self._keyring.encrypt(True, topic, args, kwargs)
+        encoded_payload = None
+        if self._payload_codec:
+            encoded_payload = self._payload_codec.encode(True, topic, args, kwargs)
 
-        if encrypted_payload:
+        if encoded_payload:
             if options:
                 msg = message.Publish(request_id,
                                       topic,
-                                      payload=encrypted_payload.payload,
-                                      enc_algo=encrypted_payload.algo,
-                                      enc_key=encrypted_payload.pkey,
-                                      enc_serializer=encrypted_payload.serializer,
+                                      payload=encoded_payload.payload,
+                                      enc_algo=encoded_payload.enc_algo,
+                                      enc_key=encoded_payload.enc_key,
+                                      enc_serializer=encoded_payload.enc_serializer,
                                       **options.message_attr())
             else:
                 msg = message.Publish(request_id,
                                       topic,
-                                      payload=encrypted_payload.payload,
-                                      enc_algo=encrypted_payload.algo,
-                                      enc_key=encrypted_payload.pkey,
-                                      enc_serializer=encrypted_payload.serializer)
+                                      payload=encoded_payload.payload,
+                                      enc_algo=encoded_payload.enc_algo,
+                                      enc_key=encoded_payload.enc_key,
+                                      enc_serializer=encoded_payload.enc_serializer)
         else:
             if options:
                 msg = message.Publish(request_id,
@@ -1220,7 +1269,7 @@ class ApplicationSession(BaseSession):
         if options and options.acknowledge:
             # only acknowledged publications expect a reply ..
             on_reply = txaio.create_future()
-            self._publish_reqs[request_id] = PublishRequest(request_id, on_reply, was_encrypted=(encrypted_payload is not None))
+            self._publish_reqs[request_id] = PublishRequest(request_id, on_reply, was_encrypted=(encoded_payload is not None))
         else:
             on_reply = None
 
@@ -1280,9 +1329,14 @@ class ApplicationSession(BaseSession):
                 if "_wampuris" in proc.__dict__:
                     for pat in proc.__dict__["_wampuris"]:
                         if pat.is_handler():
-                            uri = pat.uri()
-                            subopts = options or pat.subscribe_options()
-                            on_replies.append(_subscribe(handler, proc, uri, subopts))
+                            _uri = pat.uri()
+                            subopts = pat.options or options
+                            if subopts is None:
+                                if pat.uri_type == uri.Pattern.URI_TYPE_WILDCARD:
+                                    subopts = types.SubscribeOptions(match=u"wildcard")
+                                else:
+                                    subopts = types.SubscribeOptions(match=u"exact")
+                            on_replies.append(_subscribe(handler, proc, _uri, subopts))
 
             # XXX needs coverage
             return txaio.gather(on_replies, consume_exceptions=True)
@@ -1337,26 +1391,30 @@ class ApplicationSession(BaseSession):
 
         request_id = self._request_id_gen.next()
 
-        encrypted_payload = None
-        if self._keyring:
-            encrypted_payload = self._keyring.encrypt(True, procedure, args, kwargs)
+        encoded_payload = None
+        if self._payload_codec:
+            try:
+                encoded_payload = self._payload_codec.encode(True, procedure, args, kwargs)
+            except:
+                self.log.failure()
+                raise
 
-        if encrypted_payload:
+        if encoded_payload:
             if options:
                 msg = message.Call(request_id,
                                    procedure,
-                                   payload=encrypted_payload.payload,
-                                   enc_algo=encrypted_payload.algo,
-                                   enc_key=encrypted_payload.pkey,
-                                   enc_serializer=encrypted_payload.serializer,
+                                   payload=encoded_payload.payload,
+                                   enc_algo=encoded_payload.enc_algo,
+                                   enc_key=encoded_payload.enc_key,
+                                   enc_serializer=encoded_payload.enc_serializer,
                                    **options.message_attr())
             else:
                 msg = message.Call(request_id,
                                    procedure,
-                                   payload=encrypted_payload.payload,
-                                   enc_algo=encrypted_payload.algo,
-                                   enc_key=encrypted_payload.pkey,
-                                   enc_serializer=encrypted_payload.serializer)
+                                   payload=encoded_payload.payload,
+                                   enc_algo=encoded_payload.enc_algo,
+                                   enc_key=encoded_payload.enc_key,
+                                   enc_serializer=encoded_payload.enc_serializer)
         else:
             if options:
                 msg = message.Call(request_id,
@@ -1397,13 +1455,14 @@ class ApplicationSession(BaseSession):
         return on_reply
 
     @public
-    def register(self, endpoint, procedure=None, options=None):
+    def register(self, endpoint, procedure=None, options=None, prefix=None):
         """
         Implements :func:`autobahn.wamp.interfaces.ICallee.register`
         """
         assert((callable(endpoint) and procedure is not None) or hasattr(endpoint, '__class__'))
         assert(procedure is None or type(procedure) == six.text_type)
         assert(options is None or isinstance(options, types.RegisterOptions))
+        assert prefix is None or isinstance(prefix, six.text_type)
 
         if not self._transport:
             raise exception.TransportLost()
@@ -1412,6 +1471,8 @@ class ApplicationSession(BaseSession):
             request_id = self._request_id_gen.next()
             on_reply = txaio.create_future()
             endpoint_obj = Endpoint(fn, obj, options.details_arg if options else None)
+            if prefix is not None:
+                procedure = u"{}{}".format(prefix, procedure)
             self._register_reqs[request_id] = RegisterRequest(request_id, on_reply, procedure, endpoint_obj)
 
             if options:
@@ -1436,8 +1497,9 @@ class ApplicationSession(BaseSession):
                 if "_wampuris" in proc.__dict__:
                     for pat in proc.__dict__["_wampuris"]:
                         if pat.is_endpoint():
-                            uri = pat.uri()
-                            on_replies.append(_register(endpoint, proc, uri, options))
+                            _uri = pat.uri()
+                            regopts = pat.options or options
+                            on_replies.append(_register(endpoint, proc, _uri, regopts))
 
             # XXX neds coverage
             return txaio.gather(on_replies, consume_exceptions=True)
@@ -1464,6 +1526,131 @@ class ApplicationSession(BaseSession):
         return on_reply
 
 
+# this is NOT public; import from either autobahn.asyncio.wamp or
+# autobahn.twisted.wamp
+class _SessionShim(ApplicationSession):
+    """
+    shim that lets us present pep8 API for user-classes to override,
+    but also backwards-compatible for existing code using
+    ApplicationSession "directly".
+    """
+
+    #: name -> IAuthenticator
+    _authenticators = None
+
+    def onJoin(self, details):
+        return self.on_join(details)
+
+    def onConnect(self):
+        if self._authenticators:
+            # authid, authrole *must* match across all authenticators
+            # (checked in add_authenticator) so these lists are either
+            # [None] or [None, 'some_authid']
+            authid = [x._args.get('authid', None) for x in self._authenticators.values()][-1]
+            authrole = [x._args.get('authrole', None) for x in self._authenticators.values()][-1]
+            authextra = self._merged_authextra()
+            self.join(
+                self.config.realm,
+                authmethods=list(self._authenticators.keys()),
+                authid=authid or u'public',
+                authrole=authrole or u'default',
+                authextra=authextra,
+            )
+        else:
+            self.on_connect()
+
+    def onChallenge(self, challenge):
+        try:
+            authenticator = self._authenticators[challenge.method]
+        except KeyError:
+            raise RuntimeError(
+                "Received challenge for unknown authmethod '{}'".format(
+                    challenge.method
+                )
+            )
+        return authenticator.on_challenge(self, challenge)
+
+    def onLeave(self, details):
+        return self.on_leave(details)
+
+    def onDisconnect(self):
+        return self.on_disconnect()
+
+    # experimental authentication API
+
+    def add_authenticator(self, authenticator):
+        assert isinstance(authenticator, IAuthenticator)
+        if self._authenticators is None:
+            self._authenticators = {}
+
+        # before adding this authenticator we need to validate that
+        # it's consistent with any other authenticators we may have --
+        # for example, they must all agree on "authid" etc because
+        # .join() only takes one value for all of those.
+
+        def at_most_one(name):
+            uni = set([
+                a._args[name]
+                for a in list(self._authenticators.values()) + [authenticator]
+                if name in a._args
+            ])
+            if len(uni) > 1:
+                raise ValueError(
+                    "Inconsistent {}s: {}".format(
+                        name,
+                        ' '.join(uni),
+                    )
+                )
+
+        # all authids must match
+        at_most_one('authid')
+
+        # all authroles must match
+        at_most_one('authrole')
+
+        # can we do anything else other than merge all authextra keys?
+        # here we check that any duplicate keys have the same values
+        authextra = authenticator.authextra
+        merged = self._merged_authextra()
+        for k, v in merged:
+            if k in authextra and authextra[k] != v:
+                raise ValueError(
+                    "Inconsistent authextra values for '{}': '{}' vs '{}'".format(
+                        k, v, authextra[k],
+                    )
+                )
+
+        # validation complete, add it
+        self._authenticators[authenticator.name] = authenticator
+
+    def _merged_authextra(self):
+        authextras = [a._args.get('authextra', {}) for a in self._authenticators.values()]
+        # for all existing _authenticators, we've already checked that
+        # if they contain a key it has the same value as all others.
+        return {
+            k: v
+            for k, v in zip(
+                reduce(lambda x, y: x | set(y.keys()), authextras, set()),
+                reduce(lambda x, y: x | set(y.values()), authextras, set())
+            )
+        }
+
+    # these are the actual "new API" methods (i.e. snake_case)
+    #
+
+    def on_join(self, details):
+        pass
+
+    def on_leave(self, details):
+        self.disconnect()
+
+    def on_connect(self):
+        self.join(self.config.realm)
+
+    def on_disconnect(self):
+        pass
+
+
 # ISession.register collides with the abc.ABCMeta.register method
 # ISession.register(ApplicationSession)
 
